#include "jacobi.hh"
#include "alglin.hh"
#include <cmath>
#include <iostream>

namespace Jacobi {

  using namespace std ;

  void
  GivensLeft( indexType const n,        /* Number of rows of matrix A             */
              valueType       A[],      /* Pointer to coefficients of A           */
              indexType const lda,      /* Leading dimension of A, Fortran style  */
              indexType const row1,     /* first row for Givens rotation          */
              indexType const row2,     /* second row for Givens rotation         */
              valueType const theta ) {
    //
    //  /                        \
    //  | cos(theta)  sin(theta) |  
    //  |                        |
    //  | -sin(theta) cos(theta) |
    //  \                        /
    //
    // row1 <-  cos(theta) * row1 + sin(theta) * row2
    // row2 <- -sin(theta) * row1 + cos(theta) * row2
    valueType cs = cos(theta) ;
    valueType sn = sin(theta) ;
    valueType * R1 = A + row1 ;
    valueType * R2 = A + row2 ;
    for ( indexType i=0 ; i < n ; ++i, R1 += lda, R2 += lda ) {
      valueType tmp = cs*(*R1)+sn*(*R2) ;
      *R2 = cs*(*R2)-sn*(*R1) ;
      *R1 = tmp ;
    }  
  }

  void
  GivensRight( indexType const n,        /* Number of rows of matrix A             */
               valueType       A[],      /* Pointer to coefficients of A           */
               indexType const lda,      /* Leading dimension of A, Fortran style  */
               indexType const col1,     /* first column for Givens rotation       */
               indexType const col2,     /* second column for Givens rotation      */
               valueType const theta ) {
    //
    //  /                        \
    //  | cos(theta) -sin(theta) |  
    //  |                        |
    //  | sin(theta)  cos(theta) |
    //  \                        /
    //
    // col1 <-  cos(theta) * col1 + sin(theta) * col2
    // row2 <- -sin(theta) * col1 + cos(theta) * col2
    valueType cs = cos(theta) ;
    valueType sn = sin(theta) ;
    valueType * C1 = A + col1 * lda ;
    valueType * C2 = A + col2 * lda ;
    for ( indexType i=0 ; i < n ; ++i, ++C1, ++C2 ) {
      valueType tmp = cs*(*C1)+sn*(*C2) ;
      *C2 = cs*(*C2)-sn*(*C1) ;
      *C1 = tmp ;
    }  
  }

  valueType
  evalAngle( indexType const n,    /* Number of rows of matrix A             */
             valueType       A[],  /* Pointer to coefficients of A           */
             indexType const lda,  /* Leading dimension of A, Fortran style  */
             indexType const row,  // row > col
             indexType const col ) {
    //
    //      col
    //       |
    //    .  .  .  .  .
    //    . Aii . Aij .
    //    .  .  .  .  .
    //    .  .  .  .  .
    //    . Aij . Ajj . <--- row
    //    .  .  .  .  .
    valueType Aij = A[row + col * lda] ;
    valueType Aii = A[col + col * lda] ;
    //valueType Ajj = A[row + row * lda] ;
    return -atan2( Aij, Aii ) ;         
  }
  
  bool
  Jacobi( indexType const n,          /* Number of rows of matrix A             */
          valueType       A[],        /* Pointer to coefficients of A           */
          indexType const lda,        /* Leading dimension of A, Fortran style  */
          valueType       Q[],        /* Pointer to coefficients of Q           */
          indexType const ldq,        /* Leading dimension of Q, Fortran style  */
          valueType const epsi,       /* tolerance for off diagonal elements    */
          indexType const maxIter ) { /* maximum number of big iterations       */
    // initialize Q to the identity matrix
    valueType one = 1 ;
    alglin::zero( n*ldq, Q, 1 ) ;
    alglin::copy( n, &one, 0, Q, ldq+1 ) ;

    for ( indexType k = 0 ; k < maxIter ; ++k ) {
      // evaluate max norm of non diagonal elements
      indexType iAbs   = 0 ;
      indexType jAbs   = 0 ;
      valueType maxAbs = 0 ;
      for ( indexType i = 1 ; i < n ; ++i ) {
        for ( indexType j = 0 ; j < i ; ++j ) {
          valueType absAij = abs( A[ i + lda * j] ) ;
          if ( absAij > maxAbs ) {
            iAbs   = i ;
            jAbs   = j ;
            maxAbs = absAij ;
          }
        }
      }
      if ( maxAbs < epsi ) return true ;
      // loop for elements under diagonal
      #if 1
      // evaluate givens rotation
      valueType theta = evalAngle( n, A, lda, iAbs, jAbs ) ;
      // apply givens rotation
      GivensLeft ( n, A, lda, iAbs, jAbs, theta ) ;
      GivensRight( n, A, lda, iAbs, jAbs, theta ) ;
      // apply givens rotation to Q
      GivensRight( n, Q, ldq, iAbs, jAbs, theta ) ;      
      #else
      for ( indexType i = 1 ; i < n ; ++i ) {
        for ( indexType j = 0 ; j < i ; ++j ) {
          // evaluate givens rotation
          valueType theta = evalAngle( n, A, lda, i, j ) ;
          // apply givens rotation
          GivensLeft ( n, A, lda, i, j, theta ) ;
          GivensRight( n, A, lda, i, j, theta ) ;
          // apply givens rotation to Q
          GivensRight( n, Q, ldq, i, j, theta ) ;
        }
      }
      #endif
    }
    return false ;
  }
}
